"""
This module provides access to a class that loads data in a separate thread.
"""
import math

import numpy as np
from PySide2 import QtCore

from sohstationviewer.conf import constants as const


# Global flag that determines whether the user requested to stop processing and
# its corresponding global lock.
stop = False
stop_lock = QtCore.QMutex()


class DownsamplerWorkerSignals(QtCore.QObject):
    """
    Object that contains the signals for DownsamplerWorker.

    Signals:
        finished: emitted when a DownsamplerWorker instance finishes, contains
            the result of the computation done by the instance
        stopped: emitted when a DownsamplerWorker instance is stopped as a
            result of a request from the user
    """
    finished = QtCore.Signal(np.ndarray, np.ndarray, np.ndarray)
    stopped = QtCore.Signal()


class DownsamplerWorker(QtCore.QRunnable):
    """
    The worker class that executes the code to downsample data for plotting.
    """

    def __init__(self, times, data, log_indexes=None, rq_points=0,
                 do_downsample=True):
        super().__init__()

        # The data to be downsampled
        self.times = times
        self.data = data
        self.log_indexes = log_indexes

        # The maximum number of data points that should exist after
        # downsampling. Due to how the downsampling process works, there might
        # be more data points than this limit once processing is finished.
        self.rq_points = rq_points

        self.signals = DownsamplerWorkerSignals()

        # Whether to downsample the data. Intended for times when a user of
        # this class has some data sets that might not need downsampling.
        self.do_downsample = do_downsample

    def run(self):
        global stop
        try:
            times = self.times
            data = self.data
            log_indexes = self.log_indexes
            if self.do_downsample:
                times, data, log_indexes = downsample(
                    self.times, self.data, self.log_indexes, self.rq_points
                )
        except StopRequested:
            # The way we set things up, only one background thread can do
            # downsampling at a time. When we stop processing waveform data,
            # all other queued background downsampling tasks are killed,
            # leaving the sole running downsampler. As a result, we can reset
            # the stop flag with no issue because no other downsampler will
            # read it afterward.
            stop = False
            self.signals.stopped.emit()
        else:
            if not stop:
                self.signals.finished.emit(
                    times, data, log_indexes
                )


class Downsampler:
    """
    The class that coordinate the downsampling of data in background threads.
    Intended to be used in cases where there are multiple data sets that need
    to be downsampled. For instances where there is only one data set,
    DownsamplerWorker offers an easier-to-use API at the cost of having to
    do your own thread management.
    """

    def __init__(self):
        self.thread_pool = QtCore.QThreadPool()
        # Setting the maximum number of thread executed in the background in
        # this thread pool to 1. Doing so helps to prevent the situation where
        # multiple threads finish their work at once, which may lead to the
        # main thread hanging if the post-processing procedure is too complex.
        self.thread_pool.setMaxThreadCount(1)

        self.worker_list = []

    def add_worker(self, times, data, log_indexes=None, rq_points=0,
                   do_downsample=True):
        """
        Create a downsampler worker for a data set and add it to the worker
        list. Also returns the created worker so that users can connect its
        signals to their slots.

        :param times: the times array to downsample
        :param data: the data array to downsample
        :param log_indexes: the soh message line indices array to downsample
        :param rq_points: the requested size of the downsampled arrays
        :param do_downsample: whether to downsample the given data set. True by
            default
        :return: the worker that will downsample the given data
        """
        worker = DownsamplerWorker(times, data, log_indexes, rq_points,
                                   do_downsample)
        self.worker_list.append(worker)

        return worker

    def start(self):
        """
        Signal the internal thread pool to put all the stored downsampler
        workers into its execution queue. Each worker will begin running when
        its turn comes.
        """
        for worker in self.worker_list:
            self.thread_pool.start(worker)

    def request_stop(self):
        """
        Request the downsampler to stop all its workers by setting the global
        stop flag.
        """
        global stop, stop_lock
        stop_lock.lock()
        stop = True
        stop_lock.unlock()
        self.thread_pool.clear()


class StopRequested(Exception):
    """
    Exception that is raised to indicate that a downsampler worker should stop
    running and starts cleaning up.
    """
    pass


def downsample(times, data, log_indexes=None, rq_points=0):
    """
    Reduce sample rate of times and data so that times and data return has
        the size around the rq_points.
    Since the functions used for downsampling (chunk_minmax()/constant_rate)
        are very slow, the values of data from mean to CUT_FROM_MEAN_FACTOR
        will be removed first. If the size not meet the rq_points, then
        continue to downsample.
    :param times: numpy array - of a waveform channel's times
    :param data: numpy array - of a waveform channel's data
    :param log_indexes: numpy array - of a waveform channel's soh message line
        index
    :param rq_points: int - requested size to return.
    :return np.array, np.array,(np.array) - new times and new data (and new
        log_indexes) with the requested size
    """
    # create a dummy array for log_indexes. However this way may slow down
    # the performance of waveform downsample because waveform channel are large
    # and have no log_indexes.

    global stop, stop_lock
    stop_lock.lock()
    if stop:
        stop_lock.unlock()
        raise StopRequested()
    stop_lock.unlock()

    if times.size <= rq_points:
        return times, data, log_indexes
    if log_indexes is None:
        log_indexes = np.empty_like(times)
    data_max = max(abs(data.max()), abs(data.min()))
    data_mean = abs(data.mean())
    indexes = np.where(
        abs(data - data.mean()) >
        (data_max - data_mean) * const.CUT_FROM_MEAN_FACTOR)
    times = times[indexes]
    data = data[indexes]
    log_indexes = log_indexes[indexes]

    if times.size <= rq_points:
        return times, data, log_indexes

    return chunk_minmax(times, data, log_indexes, rq_points)


def chunk_minmax(times, data, log_indexes, rq_points):
    """
    Split data into different chunks, take the min, max of each chunk to add
        to the data return
    :param times: numpy array - of a channel's times
    :param data: numpy array - of a channel's data
    :param log_indexes: numpy array - of a channel's log_indexes
    :param rq_points: int - requested size to return.
    :return times, data: np.array, np.array - new times and new data with the
        requested size
    """
    global stop, stop_lock
    stop_lock.lock()
    if stop:
        stop_lock.unlock()
        raise StopRequested()
    stop_lock.unlock()

    final_points = 0
    if times.size <= rq_points:
        final_points += times.size
        return times, data, log_indexes

    if rq_points < 2:
        return np.empty((1, 0)), np.empty((1, 0)), np.empty((1, 0))

    # Since grabbing the min and max from each
    # chunk, need to div the requested number of points
    # by 2.
    chunk_size = rq_points // 2
    chunk_count = math.ceil(times.size / chunk_size)

    if chunk_count * chunk_size > times.size:
        chunk_count -= 1
        # Length of the trace is not divisible by the number of requested
        # points. So split into an array that is divisible by the requested
        # size, and an array that contains the excess. Downsample both,
        # and combine. This case gives slightly more samples than
        # the requested sample size, but not by much.
        times_0 = times[:chunk_count * chunk_size]
        data_0 = data[:chunk_count * chunk_size]
        log_indexes_0 = log_indexes[:chunk_count * chunk_size]

        excess_times = times[chunk_count * chunk_size:]
        excess_data = data[chunk_count * chunk_size:]
        excess_log_indexes = data[chunk_count * chunk_size:]

        new_times_0, new_data_0, new_log_indexes_0 = downsample(
            times_0, data_0, log_indexes_0, rq_points=rq_points
        )

        # right-most subarray is always smaller than
        # the initially requested number of points.
        excess_times, excess_data, excess_log_indexes = downsample(
            excess_times, excess_data, excess_log_indexes,
            rq_points=chunk_count
        )

        new_times = np.zeros(new_times_0.size + excess_times.size)
        new_data = np.zeros(new_data_0.size + excess_data.size)
        new_log_indexes = np.zeros(
            new_log_indexes_0.size + excess_log_indexes.size
        )

        new_times[:new_times_0.size] = new_times_0
        new_data[:new_data_0.size] = new_data_0
        new_log_indexes[:new_log_indexes_0.size] = new_log_indexes_0

        new_times[new_times_0.size:] = excess_times
        new_data[new_data_0.size:] = excess_data
        new_log_indexes[new_log_indexes_0.size:] = excess_log_indexes

        return new_times, new_data, new_log_indexes

    new_times = times.reshape(chunk_size, chunk_count)
    new_data = data.reshape(chunk_size, chunk_count)
    new_log_indexes = log_indexes.reshape(chunk_size, chunk_count)

    min_data_idx = np.argmin(new_data, axis=1)
    max_data_idx = np.argmax(new_data, axis=1)

    rows = np.arange(chunk_size)

    mask = np.zeros(shape=(chunk_size, chunk_count), dtype=bool)
    mask[rows, min_data_idx] = True
    mask[rows, max_data_idx] = True

    new_times = new_times[mask]
    new_data = new_data[mask]
    new_log_indexes = new_log_indexes[mask]
    return new_times, new_data, new_log_indexes


def constant_rate(times, data, rq_points):
    """
    Take sample with constant_rate regardless of the value of the data
    :param times: numpy array of a waveform channel's times
    :param data: numpy array of a waveform channel's data
    :param rq_points: requested size to return.
    :return times, data: new times and new data with the requested size
    """
    if times.size <= rq_points:
        return times, data
    rate = int(times.size/rq_points)
    if rate == 1:
        return times, data

    indexes = np.arange(0, times.size, rate)
    times = times[indexes]
    data = data[indexes]

    return times, data
